"""Test cases for JSGF parser."""
import unittest

from rhasspynlu.ini_jsgf import get_expression_count
from rhasspynlu.jsgf import (
    RuleReference,
    Sentence,
    Sequence,
    SequenceType,
    SlotReference,
    Tag,
    Word,
    walk_expression,
)


class BasicJsgfTestCase(unittest.TestCase):
    """Basic JSGF test cases."""

    def test_words(self):
        """Basic word sequence."""
        s = Sentence.parse("this is a test")
        self.assertEqual(s.items, [Word("this"), Word("is"), Word("a"), Word("test")])

    def test_optional(self):
        """Basic optional."""
        s = Sentence.parse("this is [a] test")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Word("is"),
                Sequence(
                    text="a", type=SequenceType.ALTERNATIVE, items=[Word("a"), Word("")]
                ),
                Word("test"),
            ],
        )

    def test_alternative(self):
        """Basic alternative."""
        s = Sentence.parse("this (is | a) test")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Sequence(
                    text="is | a",
                    type=SequenceType.GROUP,
                    items=[
                        Sequence(
                            text="is | a",
                            type=SequenceType.ALTERNATIVE,
                            items=[Word("is"), Word("a")],
                        )
                    ],
                ),
                Word("test"),
            ],
        )

    def test_rule_reference(self):
        """Basic rule references."""
        s = Sentence.parse("this <is-a> test")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                RuleReference(text="<is-a>", rule_name="is-a"),
                Word("test"),
            ],
        )

        s = Sentence.parse("this <is.a> test")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                RuleReference(text="<is.a>", grammar_name="is", rule_name="a"),
                Word("test"),
            ],
        )

    def test_slot_reference(self):
        """Basic slot reference."""
        s = Sentence.parse("this $is-a test")
        self.assertEqual(
            s.items,
            [Word("this"), SlotReference(text="$is-a", slot_name="is-a"), Word("test")],
        )

    def test_tag_word(self):
        """Tag a word."""
        s = Sentence.parse("this{is} a{test}")
        self.assertEqual(
            s.items,
            [Word("this", tag=Tag(tag_text="is")), Word("a", tag=Tag(tag_text="test"))],
        )

    def test_tag_group(self):
        """Tag a group."""
        s = Sentence.parse("(this is a){test}")
        group = s.items[0]
        self.assertEqual(group.tag, Tag(tag_text="test"))
        self.assertEqual(group.items, [Word("this"), Word("is"), Word("a")])

    def test_tag_alternative(self):
        """Tag an alternative."""
        s = Sentence.parse("[this is a]{test}")
        alt = s.items[0]
        self.assertEqual(alt.tag, Tag(tag_text="test"))
        self.assertEqual(
            alt.items,
            [
                Sequence(
                    text="this is a",
                    type=SequenceType.GROUP,
                    items=[Word("this"), Word("is"), Word("a")],
                ),
                Word(""),
            ],
        )

    def test_expression_count(self):
        """Test counting number of expressions."""
        s = Sentence.parse("[this] [is] [a] [test]")
        expected_count = 2 * 2 * 2 * 2
        self.assertEqual(get_expression_count(s), expected_count)

    def test_word_converters(self):
        """Test multiple converters on a single word"""
        s = Sentence.parse("this is a test!c1!c2")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Word("is"),
                Word("a"),
                Word("test", converters=["c1", "c2"]),
            ],
        )

    def test_substitution_and_converters(self):
        """Test substitution and converters on a sequence"""
        s = Sentence.parse("this (is a):test!c1!c2")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Sequence(
                    text="is a",
                    type=SequenceType.GROUP,
                    items=[Word("is"), Word("a")],
                    substitution="test",
                    converters=["c1", "c2"],
                ),
            ],
        )

    def test_entity_converters(self):
        """Test multiple converters on a tag/entity"""
        s = Sentence.parse("this is a{test!c1!c2}")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Word("is"),
                Word("a", tag=Tag(tag_text="test", converters=["c1", "c2"])),
            ],
        )

    def test_sequence_converters(self):
        """Test multiple converters on a sequence"""
        s = Sentence.parse("this (is a test)!c1!c2")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Sequence(
                    text="is a test",
                    type=SequenceType.GROUP,
                    items=[Word("is"), Word("a"), Word("test")],
                    converters=["c1", "c2"],
                ),
            ],
        )


# -----------------------------------------------------------------------------


class AdvancedJsgfTestCase(unittest.TestCase):
    """More advanced JSGF test cases."""

    def test_optional_alternative(self):
        """Combined optional and alternative."""
        s = Sentence.parse("this [is | a] test")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Sequence(
                    text="is | a",
                    type=SequenceType.ALTERNATIVE,
                    items=[
                        Sequence(
                            text="is | a",
                            type=SequenceType.ALTERNATIVE,
                            items=[Word("is"), Word("a")],
                        ),
                        Word(""),
                    ],
                ),
                Word("test"),
            ],
        )

    def test_word_substitution(self):
        """Single word substitutions."""
        s = Sentence.parse("this: :is a:test")
        self.assertEqual(
            s.items,
            [
                Word("this", substitution=""),
                Word("", substitution="is"),
                Word("a", substitution="test"),
            ],
        )

    def test_group_substitution(self):
        """Group substitution."""
        s = Sentence.parse("(this is a):test")
        group = s.items[0]
        self.assertEqual(group.substitution, "test")
        self.assertEqual(group.items, [Word("this"), Word("is"), Word("a")])

    def test_alternative_substitution(self):
        """Alternative substitution."""
        s = Sentence.parse("this [is a]:isa test")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Sequence(
                    text="is a",
                    type=SequenceType.ALTERNATIVE,
                    substitution="isa",
                    items=[
                        Sequence(
                            text="is a",
                            type=SequenceType.GROUP,
                            items=[Word("is"), Word("a")],
                        ),
                        Word(""),
                    ],
                ),
                Word("test"),
            ],
        )

    def test_tag_substitution(self):
        """Tag substitution."""
        s = Sentence.parse("(this is){a:test}")
        group = s.items[0]
        self.assertEqual(group.tag, Tag(tag_text="a", substitution="test"))
        self.assertEqual(group.items, [Word("this"), Word("is")])

    def test_tag_and_word_substitution(self):
        """Tag and word substitutions."""
        s = Sentence.parse("(this:is){a:test}")
        group = s.items[0]
        self.assertEqual(group.tag, Tag(tag_text="a", substitution="test"))
        self.assertEqual(group.items, [Word("this", substitution="is")])

    def test_nested_optionals(self):
        """Optional inside an optional."""
        s = Sentence.parse("this [[is] a] test")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Sequence(
                    text="[is] a",
                    type=SequenceType.ALTERNATIVE,
                    items=[
                        Sequence(
                            text="[is] a",
                            type=SequenceType.GROUP,
                            items=[
                                Sequence(
                                    text="is",
                                    type=SequenceType.ALTERNATIVE,
                                    items=[Word("is"), Word("")],
                                ),
                                Word("a"),
                            ],
                        ),
                        Word(""),
                    ],
                ),
                Word("test"),
            ],
        )

    def test_implicit_sequences(self):
        """Implicit sequences around alternative."""
        s = Sentence.parse("this is | a test")
        alt = s.items[0]
        self.assertEqual(alt.type, SequenceType.ALTERNATIVE)
        self.assertEqual(
            alt.items,
            [
                Sequence(
                    text="this is",
                    type=SequenceType.GROUP,
                    items=[Word("this"), Word("is")],
                ),
                Sequence(
                    text="a test",
                    type=SequenceType.GROUP,
                    items=[Word("a"), Word("test")],
                ),
            ],
        )

    def test_implicit_sequence_with_rule(self):
        """Implicit sequence around alternative with a rule reference."""
        s = Sentence.parse("this | is a <test>")
        alt = s.items[0]
        self.assertEqual(alt.type, SequenceType.ALTERNATIVE)
        self.assertEqual(
            alt.items,
            [
                Word("this"),
                Sequence(
                    text="is a <test>",
                    type=SequenceType.GROUP,
                    items=[
                        Word("is"),
                        Word("a"),
                        RuleReference(text="<test>", rule_name="test"),
                    ],
                ),
            ],
        )

    def test_tagged_group_in_optional(self):
        """Tagged group inside an optional."""
        s = Sentence.parse("this is a [($test){tag}]")
        self.assertEqual(
            s.items,
            [
                Word("this"),
                Word("is"),
                Word("a"),
                Sequence(
                    text="($test){tag}",
                    type=SequenceType.ALTERNATIVE,
                    items=[
                        Sequence(
                            text="$test",
                            type=SequenceType.GROUP,
                            tag=Tag(tag_text="tag"),
                            items=[SlotReference(slot_name="test", text="$test")],
                        ),
                        Word(""),
                    ],
                ),
            ],
        )


# -----------------------------------------------------------------------------


class OtherJsgfTestCase(unittest.TestCase):
    """Corner cases exposed by testing and users."""

    def test_den_overhead(self):
        """Sequence of optionals/alternatives plus tag substitution."""
        s = Sentence.parse(
            "toggle [the] (den | playroom) [light] {light_name:den_overhead}"
        )
        self.assertEqual(
            s.items,
            [
                Word("toggle"),
                Sequence(
                    text="the",
                    type=SequenceType.ALTERNATIVE,
                    items=[Word("the"), Word("")],
                ),
                Sequence(
                    text="den | playroom",
                    type=SequenceType.GROUP,
                    items=[
                        Sequence(
                            text="den | playroom",
                            type=SequenceType.ALTERNATIVE,
                            items=[Word("den"), Word("playroom")],
                        )
                    ],
                ),
                Sequence(
                    text="light",
                    type=SequenceType.ALTERNATIVE,
                    items=[Word("light"), Word("")],
                    tag=Tag(tag_text="light_name", substitution="den_overhead"),
                ),
            ],
        )

    def test_walk(self):
        """Test walk_expression."""
        s = Sentence.parse("set alarm for (2 | 3) minutes")

        def num2words(word):
            if not isinstance(word, Word):
                return

            try:
                n = int(word.text)
                if n == 2:
                    word.text = "two"
                    word.substitution = "2"
                elif n == 3:
                    word.text = "three"
                    word.substitution = "3"
            except ValueError:
                pass

        walk_expression(s, num2words)
        self.assertEqual(
            s.items,
            [
                Word("set"),
                Word("alarm"),
                Word("for"),
                Sequence(
                    text="2 | 3",
                    type=SequenceType.GROUP,
                    items=[
                        Sequence(
                            text="2 | 3",
                            type=SequenceType.ALTERNATIVE,
                            items=[
                                Word("two", substitution="2"),
                                Word("three", substitution="3"),
                            ],
                        )
                    ],
                ),
                Word("minutes"),
            ],
        )

    def test_walk_multiple_words(self):
        """Test walk_expression with a multi-word replacement."""
        s = Sentence.parse("set alarm for 23 minutes")

        def num2words(word):
            if not isinstance(word, Word):
                return

            try:
                n = int(word.text)
                if n == 23:
                    return Sequence(
                        text="23",
                        type=SequenceType.GROUP,
                        items=[Word("twenty"), Word("three")],
                        substitution="23",
                    )
            except ValueError:
                pass

        walk_expression(s, num2words)
        self.assertEqual(
            s.items,
            [
                Word("set"),
                Word("alarm"),
                Word("for"),
                Sequence(
                    text="23",
                    type=SequenceType.GROUP,
                    items=[Word("twenty"), Word("three")],
                    substitution="23",
                ),
                Word("minutes"),
            ],
        )

    def test_sequence_substition_in_alternative(self):
        """Test sequence substitution inside an alternative."""
        s = Sentence.parse(
            "((light one):light_1 | (light two):light_2) {name} :domain{light}"
        )
        self.assertEqual(
            s.items,
            [
                Sequence(
                    text="(light one):light_1 | (light two):light_2",
                    tag=Tag(tag_text="name"),
                    items=[
                        Sequence(
                            text="light one | light two",
                            type=SequenceType.ALTERNATIVE,
                            items=[
                                Sequence(
                                    text="light one",
                                    substitution="light_1",
                                    type=SequenceType.GROUP,
                                    items=[Word("light"), Word("one")],
                                ),
                                Sequence(
                                    text="light two",
                                    substitution="light_2",
                                    type=SequenceType.GROUP,
                                    items=[Word("light"), Word("two")],
                                ),
                            ],
                        )
                    ],
                ),
                Word(text="", substitution="domain", tag=Tag(tag_text="light")),
            ],
        )


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



"""Parsing code for ini/JSGF grammars."""
import configparser
import io
import logging
import re
import typing
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path

from .const import IntentsType, ReplacementsType, SentencesType
from .jsgf import (
    Expression,
    ParseMetadata,
    Rule,
    RuleReference,
    Sentence,
    Sequence,
    SequenceType,
    SlotReference,
    Word,
)

_LOGGER = logging.getLogger(__name__)


@dataclass
class Grammar:
    """Named JSGF grammar with rules."""

    grammar_name: str = ""
    rules: typing.List[Rule] = field(default_factory=list)

    GRAMMAR_DECLARATION = re.compile(r"^grammar ([^;]+);$")

    @classmethod
    def parse(cls, source: typing.TextIO) -> "Grammar":
        """Parse single JSGF grammar."""
        grammar = Grammar()

        # Read line-by-line
        for line in source:
            line = line.strip()
            if line.startswith("#") or (not line):
                # Skip comments/blank lines
                continue

            grammar_match = Grammar.GRAMMAR_DECLARATION.match(line)
            if grammar_match is not None:
                # grammar GrammarName;
                grammar.grammar_name = grammar_match.group(1)
            else:
                # public <RuleName> = rule body;
                # <RuleName> = rule body;
                grammar.rules.append(Rule.parse(line))  # pylint: disable=E1101

        return grammar


# -----------------------------------------------------------------------------


def parse_ini(
    source: typing.Union[str, Path, typing.TextIO],
    intent_filter: typing.Optional[typing.Callable[[str], bool]] = None,
    sentence_transform: typing.Callable[[str], str] = None,
    file_name: typing.Optional[str] = None,
) -> IntentsType:
    """Parse multiple JSGF grammars from an ini file."""
    intent_filter = intent_filter or (lambda x: True)
    if isinstance(source, str):
        source = io.StringIO(source)
        file_name = file_name or "<StringIO>"
    elif isinstance(source, Path):
        source = open(source, "r")
        file_name = file_name or str(source)
    else:
        file_name = file_name or "<TextIO>"

    # Process configuration sections
    sentences: IntentsType = defaultdict(list)

    try:
        # Create ini parser
        config = configparser.ConfigParser(
            allow_no_value=True, strict=False, delimiters=["="]
        )

        # case sensitive
        config.optionxform = str  # type: ignore
        config.read_file(source)

        _LOGGER.debug("Loaded ini file")

        # Parse each section (intent)
        line_number: int = 1
        for sec_name in config.sections():
            # Exclude if filtered out.
            if not intent_filter(sec_name):
                _LOGGER.debug("Skipping %s", sec_name)
                continue

            # Section header
            line_number += 1

            # Processs settings (sentences/rules)
            for k, v in config[sec_name].items():
                if v is None:
                    # Collect non-valued keys as sentences
                    sentence = k.strip()

                    # Fix \[ escape sequence
                    sentence = sentence.replace("\\[", "[")

                    if sentence_transform:
                        # Do transform
                        sentence = sentence_transform(sentence)

                    sentences[sec_name].append(
                        Sentence.parse(
                            sentence,
                            metadata=ParseMetadata(
                                file_name=file_name,
                                line_number=line_number,
                                intent_name=sec_name,
                            ),
                        )
                    )
                else:
                    sentence = v.strip()

                    if sentence_transform:
                        # Do transform
                        sentence = sentence_transform(sentence)

                    # Collect key/value pairs as JSGF rules
                    rule = f"<{k.strip()}> = ({sentence});"

                    # Fix \[ escape sequence
                    rule = rule.replace("\\[", "[")

                    sentences[sec_name].append(
                        Rule.parse(
                            rule,
                            metadata=ParseMetadata(
                                file_name=file_name,
                                line_number=line_number,
                                intent_name=sec_name,
                            ),
                        )
                    )

                # Sentence
                line_number += 1

            # Blank line
            line_number += 1
    finally:
        source.close()

    return sentences


# -----------------------------------------------------------------------------


def split_rules(
    intents: IntentsType, replacements: typing.Optional[ReplacementsType] = None
) -> typing.Tuple[SentencesType, ReplacementsType]:
    """Seperate out rules and sentences from all intents."""
    sentences: SentencesType = {}
    replacements = replacements or {}

    for intent_name, intent_exprs in intents.items():
        sentences[intent_name] = []

        # Extract rules and fold them into replacements
        for expr in intent_exprs:
            if isinstance(expr, Rule):
                # Rule
                rule_name = expr.rule_name

                # Surround with <>
                rule_name = f"<{intent_name}.{rule_name}>"
                replacements[rule_name] = [expr.rule_body]
            else:
                sentences[intent_name].append(expr)

    return sentences, replacements


# -----------------------------------------------------------------------------


def get_intent_counts(
    sentences: SentencesType,
    replacements: typing.Optional[ReplacementsType] = None,
    exclude_slots: bool = True,
    count_dict: typing.Optional[typing.Dict[Expression, int]] = None,
) -> typing.Dict[str, int]:
    """Get number of possible sentences for each intent."""
    intent_counts: typing.Dict[str, int] = defaultdict(int)

    for intent_name, intent_sentences in sentences.items():
        # Compute counts for all sentences
        intent_counts[intent_name] = max(
            1,
            sum(
                get_expression_count(
                    s, replacements, exclude_slots=exclude_slots, count_dict=count_dict
                )
                for s in intent_sentences
            ),
        )

    return intent_counts


# -----------------------------------------------------------------------------


def get_expression_count(
    expression: Expression,
    replacements: typing.Optional[ReplacementsType] = None,
    exclude_slots: bool = True,
    count_dict: typing.Optional[typing.Dict[Expression, int]] = None,
) -> int:
    """Get the number of possible sentences in an expression."""
    if isinstance(expression, Sequence):
        if expression.type == SequenceType.GROUP:
            # Counts multiply down the sequence
            count = 1
            for sub_item in expression.items:
                count = count * get_expression_count(
                    sub_item,
                    replacements,
                    exclude_slots=exclude_slots,
                    count_dict=count_dict,
                )

            if count_dict is not None:
                count_dict[expression] = count

            return count

        if expression.type == SequenceType.ALTERNATIVE:
            # Counts sum across the alternatives
            count = sum(
                get_expression_count(
                    sub_item,
                    replacements,
                    exclude_slots=exclude_slots,
                    count_dict=count_dict,
                )
                for sub_item in expression.items
            )

            if count_dict is not None:
                count_dict[expression] = count

            return count
    elif isinstance(expression, RuleReference):
        # Get substituted sentences for <rule>
        key = f"<{expression.full_rule_name}>"
        assert replacements, key
        count = sum(
            get_expression_count(
                value, replacements, exclude_slots=exclude_slots, count_dict=count_dict
            )
            for value in replacements[key]
        )

        if count_dict is not None:
            count_dict[expression] = count

        return count
    elif (not exclude_slots) and isinstance(expression, SlotReference):
        # Get substituted sentences for $slot
        key = f"${expression.slot_name}"
        assert replacements, key
        count = sum(
            get_expression_count(
                value, replacements, exclude_slots=exclude_slots, count_dict=count_dict
            )
            for value in replacements[key]
        )

        if count_dict is not None:
            count_dict[expression] = count

        return count
    elif isinstance(expression, Word):
        # Single word
        count = 1
        if count_dict is not None:
            count_dict[expression] = count

        return count

    # Unknown expression type
    count = 0
    if count_dict is not None:
        count_dict[expression] = count

    return count



"""Utilities for Rhasspy natural language understanding."""
from .arpa_lm import arpa_to_fst, fst_to_arpa, graph_to_arpa, graph_to_arpa_small
from .evaluate import evaluate_intents
from .fsticuffs import recognize
from .g2p import PronunciationsType, guess_pronunciations, read_pronunciations
from .ini_jsgf import parse_ini
from .jsgf import Rule, Sentence
from .jsgf_graph import (
    graph_to_fst,
    graph_to_gzip_pickle,
    graph_to_json,
    gzip_pickle_to_graph,
    intents_to_graph,
    json_to_graph,
    sentences_to_graph,
)
from .ngram import get_intent_ngram_counts
from .numbers import number_range_transform, number_transform, replace_numbers
from .slots import get_slot_replacements



"""Recognition functions for sentences using JSGF graphs."""
import base64
import itertools
import random
import time
import typing
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime

import networkx as nx

from .intent import Entity, Intent, Recognition, RecognitionResult
from .jsgf_graph import get_start_end_nodes
from .utils import pairwise

# -----------------------------------------------------------------------------

PathNodeType = typing.Union[int, typing.Tuple[int, typing.List[str]]]
PathType = typing.List[PathNodeType]


def recognize(
    tokens: typing.Union[str, typing.List[str]],
    graph: nx.DiGraph,
    fuzzy: bool = True,
    stop_words: typing.Optional[typing.Set[str]] = None,
    intent_filter: typing.Optional[typing.Callable[[str], bool]] = None,
    word_transform: typing.Optional[typing.Callable[[str], str]] = None,
    converters: typing.Optional[
        typing.Dict[str, typing.Callable[..., typing.Any]]
    ] = None,
    extra_converters: typing.Optional[
        typing.Dict[str, typing.Callable[..., typing.Any]]
    ] = None,
    **search_args,
) -> typing.List[Recognition]:
    """Recognize one or more intents from tokens or a sentence."""
    start_time = time.perf_counter()

    if isinstance(tokens, str):
        # Assume whitespace separation
        tokens = tokens.split()

    if fuzzy:
        # Fuzzy recognition
        best_fuzzy = best_fuzzy_cost(
            paths_fuzzy(
                tokens,
                graph,
                stop_words=stop_words,
                intent_filter=intent_filter,
                word_transform=word_transform,
                **search_args,
            )
        )

        end_time = time.perf_counter()

        if best_fuzzy:
            recognitions = []

            # Gather all successful fuzzy paths
            for fuzzy_result in best_fuzzy:
                result, recognition = path_to_recognition(
                    fuzzy_result.node_path,
                    graph,
                    cost=fuzzy_result.cost,
                    converters=converters,
                    extra_converters=extra_converters,
                )
                if result == RecognitionResult.SUCCESS:
                    assert recognition is not None
                    recognition.recognize_seconds = end_time - start_time
                    recognitions.append(recognition)

            return recognitions
    else:
        # Strict recognition
        paths = list(
            paths_strict(
                tokens,
                graph,
                intent_filter=intent_filter,
                word_transform=word_transform,
                **search_args,
            )
        )
        if (not paths) and stop_words:
            # Try again by excluding stop words
            tokens = [t for t in tokens if t not in stop_words]
            paths = list(
                paths_strict(
                    tokens,
                    graph,
                    exclude_tokens=stop_words,
                    intent_filter=intent_filter,
                    word_transform=word_transform,
                    **search_args,
                )
            )

        end_time = time.perf_counter()
        recognitions = []
        for path in paths:
            result, recognition = path_to_recognition(
                path, graph, converters=converters, extra_converters=extra_converters
            )
            if result == RecognitionResult.SUCCESS:
                assert recognition is not None
                recognition.recognize_seconds = end_time - start_time
                recognitions.append(recognition)

        return recognitions

    # No results
    return []


# -----------------------------------------------------------------------------


def paths_strict(
    tokens: typing.List[str],
    graph: nx.DiGraph,
    exclude_tokens: typing.Optional[typing.Set[str]] = None,
    max_paths: typing.Optional[int] = None,
    intent_filter: typing.Optional[typing.Callable[[str], bool]] = None,
    word_transform: typing.Optional[typing.Callable[[str], str]] = None,
) -> typing.Iterable[PathType]:
    """Match a single path from the graph exactly if possible."""
    if not tokens:
        return []

    intent_filter = intent_filter or (lambda x: True)
    word_transform = word_transform or (lambda x: x)

    # node -> attrs
    n_data = graph.nodes(data=True)

    # start state
    start_node, _ = get_start_end_nodes(graph)
    assert start_node is not None

    # Number of matching paths found
    paths_found: int = 0

    # Do breadth-first search.
    # Queue contains items of the form:
    # * current node (int)
    # * current path (int, str?) - node, matching input token
    # * remaining input tokens
    node_queue: typing.List[typing.Tuple[int, PathType, typing.List[str]]] = [
        (start_node, [], tokens)
    ]

    while node_queue:
        current_node, current_path, current_tokens = node_queue.pop(0)
        is_final = n_data[current_node].get("final", False)
        if is_final and (not current_tokens):
            # Reached final state
            paths_found += 1
            yield current_path

            if max_paths and (paths_found >= max_paths):
                break

        for next_node, edge_data in graph[current_node].items():
            next_path = list(current_path)
            next_tokens = list(current_tokens)

            ilabel = edge_data.get("ilabel", "")
            olabel = edge_data.get("olabel", "")
            matching_tokens: typing.List[str] = []

            if olabel.startswith("__label__"):
                intent_name = olabel[9:]
                if not intent_filter(intent_name):
                    # Skip intent
                    continue

            if ilabel:
                ilabel = word_transform(ilabel)

                if next_tokens:
                    # Failed to match input label
                    if ilabel != word_transform(next_tokens[0]):
                        if (not exclude_tokens) or (ilabel not in exclude_tokens):
                            # Can't exclude
                            continue
                    else:
                        # Token match
                        matching_token = next_tokens.pop(0)
                        matching_tokens.append(matching_token)
                else:
                    # Ran out of tokens
                    continue

            next_path.append((current_node, matching_tokens))

            # Continue search
            node_queue.append((next_node, next_path, next_tokens))

    # No results
    return []


# -----------------------------------------------------------------------------


@dataclass
class FuzzyResult:
    """Single path for fuzzy recognition."""

    intent_name: str
    node_path: PathType
    cost: float


@dataclass
class FuzzyCostInput:
    """Input to fuzzy cost function."""

    ilabel: str
    tokens: typing.List[str]
    stop_words: typing.Set[str]
    word_transform: typing.Optional[typing.Callable[[str], str]] = None


@dataclass
class FuzzyCostOutput:
    """Output from fuzzy cost function."""

    cost: float
    continue_search: bool = True
    matching_tokens: typing.List[str] = field(default_factory=list)


def default_fuzzy_cost(cost_input: FuzzyCostInput) -> FuzzyCostOutput:
    """Increases cost when input tokens fail to match graph. Marginal cost for stop words."""
    ilabel = cost_input.ilabel
    cost = 0.0
    tokens = cost_input.tokens
    stop_words = cost_input.stop_words
    word_transform = cost_input.word_transform or (lambda x: x)
    matching_tokens: typing.List[str] = []

    if ilabel:
        ilabel = word_transform(ilabel)
        while tokens and (ilabel != word_transform(tokens[0])):
            bad_token = word_transform(tokens.pop(0))
            if bad_token in stop_words:
                # Marginal cost to ensure paths matching stop words are preferred
                cost += 0.1
            else:
                # Mismatched token
                cost += 1

        if tokens and (ilabel == word_transform(tokens[0])):
            # Consume matching token
            matching_token = tokens.pop(0)
            matching_tokens.append(matching_token)
        else:
            # No matching token
            return FuzzyCostOutput(cost=cost, continue_search=False)

    return FuzzyCostOutput(cost=cost, matching_tokens=matching_tokens)


def paths_fuzzy(
    tokens: typing.List[str],
    graph: nx.DiGraph,
    stop_words: typing.Optional[typing.Set[str]] = None,
    cost_function: typing.Optional[
        typing.Callable[[FuzzyCostInput], FuzzyCostOutput]
    ] = None,
    intent_filter: typing.Optional[typing.Callable[[str], bool]] = None,
    word_transform: typing.Optional[typing.Callable[[str], str]] = None,
) -> typing.Dict[str, typing.List[FuzzyResult]]:
    """Do less strict matching using a cost function and optional stop words."""
    if not tokens:
        return {}

    intent_filter = intent_filter or (lambda x: True)
    cost_function = cost_function or default_fuzzy_cost
    stop_words = stop_words or set()

    # node -> attrs
    n_data = graph.nodes(data=True)

    # start state
    start_node: int = next(n for n, data in n_data if data.get("start"))

    # intent -> [(symbols, cost), (symbols, cost)...]
    intent_symbols_and_costs: typing.Dict[str, typing.List[FuzzyResult]] = defaultdict(
        list
    )

    # Lowest cost so far
    best_cost: float = float(len(n_data))

    # (node, in_tokens, out_path, out_count, cost, intent_name)
    node_queue: typing.List[
        typing.Tuple[int, typing.List[str], PathType, int, float, typing.Optional[str]]
    ] = [(start_node, tokens, [], 0, 0.0, None)]

    # BFS it up
    while node_queue:
        (
            q_node,
            q_in_tokens,
            q_out_path,
            q_out_count,
            q_cost,
            q_intent,
        ) = node_queue.pop(0)
        is_final: bool = n_data[q_node].get("final", False)

        # Update best intent cost on final state.
        # Don't bother reporting intents that failed to consume any tokens.
        if is_final and (q_cost < q_out_count):
            q_intent = q_intent or ""
            best_intent_cost: typing.Optional[float] = None
            best_intent_costs = intent_symbols_and_costs.get(q_intent)
            if best_intent_costs:
                best_intent_cost = best_intent_costs[0].cost

            final_cost = q_cost + len(q_in_tokens)  # remaning tokens count against
            final_path = q_out_path

            if (best_intent_cost is None) or (final_cost < best_intent_cost):
                # Overwrite best cost
                intent_symbols_and_costs[q_intent] = [
                    FuzzyResult(
                        intent_name=q_intent, node_path=final_path, cost=final_cost
                    )
                ]
            elif final_cost == best_intent_cost:
                # Add to existing list
                intent_symbols_and_costs[q_intent].append(
                    (
                        FuzzyResult(
                            intent_name=q_intent, node_path=final_path, cost=final_cost
                        )
                    )
                )

            if final_cost < best_cost:
                # Update best cost so far
                best_cost = final_cost

        if q_cost > best_cost:
            # Can't get any better
            continue

        # Process child edges
        for next_node, edge_data in graph[q_node].items():
            in_label = edge_data.get("ilabel") or ""
            out_label = edge_data.get("olabel") or ""
            next_in_tokens = list(q_in_tokens)
            next_out_path = list(q_out_path)
            next_out_count = q_out_count
            next_cost = q_cost
            next_intent = q_intent

            if out_label:
                if out_label.startswith("__label__"):
                    next_intent = out_label[9:]
                    if not intent_filter(next_intent):
                        # Skip intent
                        continue
                elif not out_label.startswith("__"):
                    next_out_count += 1

            cost_output = cost_function(
                FuzzyCostInput(
                    ilabel=in_label,
                    tokens=next_in_tokens,
                    stop_words=stop_words,
                    word_transform=word_transform,
                )
            )

            next_cost += cost_output.cost

            if not cost_output.continue_search:
                continue

            # Extend current path
            next_out_path.append((q_node, cost_output.matching_tokens))

            node_queue.append(
                (
                    next_node,
                    next_in_tokens,
                    next_out_path,
                    next_out_count,
                    next_cost,
                    next_intent,
                )
            )

    return intent_symbols_and_costs


def best_fuzzy_cost(
    intent_symbols_and_costs: typing.Dict[str, typing.List[FuzzyResult]]
) -> typing.List[FuzzyResult]:
    """Return fuzzy results with cost."""
    if not intent_symbols_and_costs:
        return []

    best_cost: typing.Optional[float] = None
    best_results: typing.List[FuzzyResult] = []

    # Find all results with the lowest cost
    for fuzzy_results in intent_symbols_and_costs.values():
        if not fuzzy_results:
            continue

        # All results for a given intent should have the same cost
        if best_cost is None:
            # First result
            best_cost = fuzzy_results[0].cost
            best_results = list(fuzzy_results)
        elif fuzzy_results[0].cost < best_cost:
            # Overwrite
            best_results = list(fuzzy_results)
        elif fuzzy_results[0].cost == best_cost:
            # Add to list
            best_results.extend(fuzzy_results)

    return best_results


# -----------------------------------------------------------------------------


@dataclass
class ConverterInfo:
    """Local info for converter stack in path_to_recognition"""

    # Name + args
    key: str

    # Name of converter
    name: str

    # Optional arguments passed using name,arg1,arg2,...
    args: typing.Optional[typing.List[str]] = None

    # List of raw/substituted tokens
    tokens: typing.List[typing.Tuple[str, str, typing.List[str]]] = field(
        default_factory=list
    )


def path_to_recognition(
    node_path: typing.Iterable[PathNodeType],
    graph: nx.DiGraph,
    cost: typing.Optional[float] = None,
    converters: typing.Optional[
        typing.Dict[str, typing.Callable[..., typing.Any]]
    ] = None,
    extra_converters: typing.Optional[
        typing.Dict[str, typing.Callable[..., typing.Any]]
    ] = None,
) -> typing.Tuple[RecognitionResult, typing.Optional[Recognition]]:
    """Transform node path in graph to an intent recognition object."""
    if not node_path:
        # Empty path indicates failure
        return RecognitionResult.FAILURE, None

    converters = converters or get_default_converters()
    if extra_converters:
        # Merge in extra converters
        converters.update(extra_converters)

    node_attrs = graph.nodes(data=True)
    recognition = Recognition(intent=Intent("", confidence=1.0))

    # Step 1: go through path pairwise and gather input/output labels
    raw_sub_tokens: typing.List[typing.Tuple[str, str, typing.List[str]]] = []

    for last_node_tokens, next_node_tokens in pairwise(node_path):
        # Unpack path nodes
        if isinstance(last_node_tokens, int):
            last_node = last_node_tokens
            last_tokens = []
        else:
            last_node, last_tokens = last_node_tokens

        if isinstance(next_node_tokens, int):
            next_node = next_node_tokens
        else:
            next_node, _ = next_node_tokens

        # Get raw text
        word = node_attrs[next_node].get("word") or ""

        # Get output label
        edge_data = graph[last_node][next_node]
        olabel = edge_data.get("olabel") or ""

        if olabel.startswith("__unpack__"):
            # Decode payload as base64-encoded bytestring
            olabel = base64.decodebytes(olabel[10:].encode()).decode()

        if olabel.startswith("__label__"):
            # Intent name
            assert recognition.intent is not None
            recognition.intent.name = olabel[9:]
        elif word or olabel:
            # Keep non-empty words
            raw_sub_tokens.append((word, olabel, last_tokens))

    # Step 2: apply converters
    converter_stack: typing.List[ConverterInfo] = []
    raw_conv_tokens: typing.List[typing.Tuple[str, typing.Any, typing.List[str]]] = []

    for raw_token, sub_token, original_tokens in raw_sub_tokens:
        if sub_token and converter_stack and (not sub_token.startswith("__")):
            # Add to existing converter
            converter_stack[-1].tokens.append((raw_token, sub_token, original_tokens))
        elif sub_token.startswith("__convert__"):
            # Begin converter
            converter_key = sub_token[11:]
            converter_name = converter_key
            converter_args: typing.Optional[typing.List[str]] = None

            # Detect arguments
            if "," in converter_name:
                converter_name, *converter_args = converter_name.split(",")

            converter_stack.append(
                ConverterInfo(
                    key=converter_key, name=converter_name, args=converter_args
                )
            )
        elif sub_token.startswith("__converted__"):
            # End converter
            assert converter_stack, "Found __converted__ without a __convert__"
            last_converter = converter_stack.pop()
            actual_key = sub_token[13:]
            assert (
                last_converter.key == actual_key
            ), f"Mismatched converter name (expected {last_converter.key}, got {actual_key})"

            # Convert and add directly
            raw_tokens: typing.List[str] = []
            sub_tokens: typing.List[str] = []
            orig_tokens: typing.List[typing.List[str]] = []

            for t in last_converter.tokens:
                if t[0]:
                    raw_tokens.append(t[0])

                if t[1]:
                    sub_tokens.append(t[1])

                if t[2]:
                    orig_tokens.append(t[2])

            # Run substituted tokens through conversion function
            converter_func = converters[last_converter.name]

            # Pass arguments as keyword "converter_args"
            converter_kwargs = (
                {"converter_args": last_converter.args} if last_converter.args else {}
            )
            converted_tokens = converter_func(*sub_tokens, **converter_kwargs)

            if converter_stack:
                # Add to parent converter
                target_list = converter_stack[-1].tokens
            else:
                # Add directly to list
                target_list = raw_conv_tokens

            # Zip 'em up
            target_list.extend(
                itertools.zip_longest(
                    raw_tokens, converted_tokens, orig_tokens, fillvalue=""
                )
            )
        else:
            raw_conv_tokens.append((raw_token, sub_token, original_tokens))

    assert not converter_stack, f"Converter(s) remaining on stack ({converter_stack})"

    # Step 3: collect entities
    entity_stack: typing.List[Entity] = []
    raw_index = 0
    sub_index = 0

    for raw_token, conv_token, original_tokens in raw_conv_tokens:
        # Handle raw (input) token
        if original_tokens:
            # Use tokens from recognition string
            recognition.raw_tokens.extend(original_tokens)
            raw_index += sum(len(t) for t in original_tokens) + len(original_tokens)

            if entity_stack:
                last_entity = entity_stack[-1]
                last_entity.raw_tokens.extend(original_tokens)
        elif raw_token:
            # Use word itself
            recognition.raw_tokens.append(raw_token)
            raw_index += len(raw_token) + 1

            if entity_stack:
                last_entity = entity_stack[-1]
                last_entity.raw_tokens.append(raw_token)

        # Handle converted (output) token
        if conv_token is not None:
            conv_token_str = str(conv_token)
            if conv_token_str:
                if conv_token_str.startswith("__begin__"):
                    # Begin tag/entity
                    entity_name = conv_token[9:]
                    entity_stack.append(
                        Entity(
                            entity=entity_name,
                            value="",
                            start=sub_index,
                            raw_start=raw_index,
                        )
                    )
                elif conv_token_str.startswith("__end__"):
                    # End tag/entity
                    assert entity_stack, "Found __end__ without a __begin__"
                    last_entity = entity_stack.pop()
                    actual_name = conv_token[7:]
                    assert (
                        last_entity.entity == actual_name
                    ), "Mismatched entity name (expected {last_entity.entity}, got {actual_name})"

                    # Assign end indexes
                    last_entity.end = sub_index - 1
                    last_entity.raw_end = raw_index - 1

                    # Create values
                    if len(last_entity.tokens) == 1:
                        # Use Python object
                        last_entity.value = last_entity.tokens[0]
                    else:
                        # Join as string
                        last_entity.value = " ".join(str(t) for t in last_entity.tokens)

                    last_entity.raw_value = " ".join(last_entity.raw_tokens)

                    # Add to recognition
                    recognition.entities.append(last_entity)
                elif conv_token_str.startswith("__source__"):
                    if entity_stack:
                        last_entity = entity_stack[-1]
                        last_entity.source = conv_token_str[10:]
                elif entity_stack:
                    # Add to most recent named entity
                    last_entity = entity_stack[-1]
                    last_entity.tokens.append(conv_token)

                    recognition.tokens.append(conv_token)
                    sub_index += len(conv_token_str) + 1
                else:
                    # Substituted text
                    recognition.tokens.append(conv_token)
                    sub_index += len(conv_token_str) + 1

    # Step 4: create text fields and compute confidence
    recognition.text = " ".join(str(t) for t in recognition.tokens)
    recognition.raw_text = " ".join(recognition.raw_tokens)

    if cost and cost > 0:
        # Set fuzzy confidence
        assert recognition.intent is not None

        recognition.intent.confidence = 1.0 - float(cost / len(recognition.raw_tokens))

    return RecognitionResult.SUCCESS, recognition


# -----------------------------------------------------------------------------


def bool_converter(obj: typing.Any) -> bool:
    """Returns False if argument is 'false' string or zero."""
    return (obj != 0) and (str(obj).lower() != "false")


def get_default_converters() -> typing.Dict[str, typing.Callable[..., typing.Any]]:
    """Get built-in fsticuffs converters"""
    return {
        "int": lambda *args: map(int, args),
        "float": lambda *args: map(float, args),
        "bool": lambda *args: map(bool_converter, args),
        "lower": lambda *args: map(str.lower, args),
        "upper": lambda *args: map(str.upper, args),
        "object": lambda *args: [
            {"value": args[0] if len(args) == 1 else " ".join(str(a) for a in args)}
        ],
        "kind": lambda *args, converter_args=None: [
            {"kind": converter_args[0], **a} for a in args
        ],
        "unit": lambda *args, converter_args=None: [
            {"unit": converter_args[0], **a} for a in args
        ],
        "datetime": lambda *args, converter_args=None: [
            datetime.strptime(" ".join(str(a) for a in args), *converter_args)
        ],
    }


# -----------------------------------------------------------------------------


def sample_by_intent(
    intent_graph: nx.DiGraph,
    num_samples: typing.Optional[int] = None,
    **recognition_args,
) -> typing.Dict[str, typing.List[Recognition]]:
    """Sample sentences from a graph."""

    sentences_by_intent: typing.Dict[str, typing.List[Recognition]] = defaultdict(list)

    start_node, end_node = get_start_end_nodes(intent_graph)
    assert (start_node is not None) and (
        end_node is not None
    ), "Missing start/end node(s)"

    if num_samples is not None:
        # Randomly sample
        paths = random.sample(
            list(nx.all_simple_paths(intent_graph, start_node, end_node)), num_samples
        )
    else:
        # Use generator
        paths = nx.all_simple_paths(intent_graph, start_node, end_node)

    for path in paths:
        _, recognition = path_to_recognition(path, intent_graph, **recognition_args)
        assert recognition, "Path failed"
        if recognition.intent:
            sentences_by_intent[recognition.intent.name].append(recognition)

    return sentences_by_intent



"""Parses a subset of JSGF into objects."""
import re
import typing
from dataclasses import dataclass, field
from enum import Enum


@dataclass
class Substitutable:
    """Indicates an expression may be replaced with some text."""

    # Replacement text
    substitution: typing.Optional[typing.Union[str, typing.List[str]]] = None

    # Names of converters to apply after substitution
    converters: typing.List[str] = field(default_factory=list)

    @classmethod
    def parse_substitution(cls, sub_text: str) -> typing.Union[str, typing.List[str]]:
        """Parse substitution text into token list or string."""
        sub_text = sub_text.strip()

        if sub_text.startswith("("):
            sub_text = sub_text[1:]

        if sub_text.endswith(")"):
            sub_text = sub_text[:-1]

        if " " in sub_text:
            return sub_text.split()

        return sub_text


@dataclass
class Tag(Substitutable):
    """{tag} attached to an expression."""

    # Name of tag (entity)
    tag_text: str = ""


@dataclass
class Taggable:
    """Indicates an expression may be tagged."""

    # Tag to be applied
    tag: typing.Optional[Tag] = None


@dataclass
class Expression:
    """Base class for most JSGF types."""

    # Text representation expression
    text: str = ""


@dataclass
class Word(Substitutable, Taggable, Expression):
    """Single word/token."""


class SequenceType(str, Enum):
    """Type of a sequence. Optionals are alternatives with an empty option."""

    # Sequence of expressions
    GROUP = "group"

    # Expressions where only one will be recognized
    ALTERNATIVE = "alternative"


@dataclass
class Sequence(Substitutable, Taggable, Expression):
    """Ordered sequence of expressions. Supports groups, optionals, and alternatives."""

    # Items in the sequence
    items: typing.List[Expression] = field(default_factory=list)

    # Group or alternative
    type: SequenceType = SequenceType.GROUP


@dataclass
class RuleReference(Taggable, Expression):
    """Reference to a rule by <name> or <grammar.name>."""

    # Name of referenced rule
    rule_name: str = ""

    # Grammar name of referenced rule
    grammar_name: typing.Optional[str] = None

    @property
    def full_rule_name(self):
        """Get fully qualified rule name."""
        if self.grammar_name:
            return f"{self.grammar_name}.{self.rule_name}"

        return self.rule_name


@dataclass
class SlotReference(Substitutable, Taggable, Expression):
    """Reference to a slot by $name."""

    # Name of referenced slot
    slot_name: str = ""


@dataclass
class ParseMetadata:
    """Debug metadata for more helpful parsing errors."""

    file_name: str
    line_number: int
    intent_name: typing.Optional[str] = None


@dataclass
class Sentence(Sequence):
    """Sequence representing a complete sentence template."""

    @classmethod
    def parse(
        cls, text: str, metadata: typing.Optional[ParseMetadata] = None
    ) -> "Sentence":
        """Parse a single sentence."""
        s = Sentence(text=text)
        parse_expression(s, text, metadata=metadata)
        return Sentence(
            text=s.text,
            items=s.items,
            type=s.type,
            tag=s.tag,
            substitution=s.substitution,
        )


@dataclass
class Rule:
    """Named rule with body."""

    RULE_DEFINITION = re.compile(r"^(public)?\s*<([^>]+)>\s*=\s*([^;]+)(;)?$")

    rule_name: str
    rule_body: Sentence
    public: bool = False
    text: str = ""

    @classmethod
    def parse(
        cls, text: str, metadata: typing.Optional[ParseMetadata] = None
    ) -> "Rule":
        """Parse a single rule."""
        # public <RuleName> = rule body;
        # <RuleName> = rule body;
        rule_match = Rule.RULE_DEFINITION.match(text)
        assert rule_match is not None, f"No rule was found in {text}"

        public = rule_match.group(1) is not None
        rule_name = rule_match.group(2)
        rule_text = rule_match.group(3)

        s = Sentence.parse(rule_text, metadata=metadata)
        return Rule(rule_name=rule_name, rule_body=s, public=public, text=text)


# -----------------------------------------------------------------------------


def walk_expression(
    expression: Expression,
    visit: typing.Callable[
        [Expression], typing.Union[bool, typing.Optional[Expression]]
    ],
    replacements: typing.Optional[typing.Dict[str, typing.List[Expression]]] = None,
) -> typing.Union[bool, typing.Optional[Expression]]:
    """Recursively visit/replace nodes in expression."""
    result = visit(expression)

    if result is False:
        return False

    if result is not None:
        assert isinstance(result, Expression), f"Expected Expression, got {result}"
        expression = result

    if isinstance(expression, Sequence):
        for i in range(len(expression.items)):
            new_item = walk_expression(expression.items[i], visit, replacements)
            if new_item:
                assert isinstance(
                    new_item, Expression
                ), f"Expected Expression, got {new_item}"
                expression.items[i] = new_item
    elif isinstance(expression, Rule):
        new_body = walk_expression(expression.rule_body, visit, replacements)
        if new_body:
            assert isinstance(new_body, Sentence), f"Expected Sentence, got {new_body}"
            expression.rule_body = new_body
    elif isinstance(expression, RuleReference):
        key = f"<{expression.full_rule_name}>"
        if replacements and (key in replacements):
            key_replacements = replacements[key]
            for i in range(len(key_replacements)):
                new_item = walk_expression(key_replacements[i], visit, replacements)
                if new_item:
                    assert isinstance(
                        new_item, Expression
                    ), f"Expected Expression, got {new_item}"
                    key_replacements[i] = new_item
    elif isinstance(expression, SlotReference):
        key = f"${expression.slot_name}"
        if replacements and (key in replacements):
            key_replacements = replacements[key]
            for i in range(len(key_replacements)):
                new_item = walk_expression(key_replacements[i], visit, replacements)
                if new_item:
                    assert isinstance(
                        new_item, Expression
                    ), f"Expected Expression, got {new_item}"
                    key_replacements[i] = new_item

    return expression


# -----------------------------------------------------------------------------


def split_words(text: str) -> typing.Iterable[Expression]:
    """Split words by whitespace. Detect slot references and substitutions."""
    tokens: typing.List[str] = []
    token: str = ""
    last_c: str = ""
    in_seq_sub: bool = False

    # Process words, correctly handling substitution sequences.
    # e.g., input:(output words)
    for c in text:
        break_token = False

        if (c == "(") and (last_c == ":"):
            # Begin sequence substitution
            in_seq_sub = True
        elif in_seq_sub and (c == ")"):
            # End sequence substitution
            in_seq_sub = False
            break_token = True
        elif c == " " and (not in_seq_sub):
            # Whitespace break (not inside sequence substitution)
            break_token = True
        else:
            # Accumulate into token
            token += c

        if break_token and token:
            tokens.append(token)
            token = ""

        last_c = c

    if token:
        # Last token
        tokens.append(token)

    for token in tokens:
        if token.startswith("$"):
            if ":" in token:
                # Slot with substitutions
                lhs, rhs = token[1:].split(":", maxsplit=1)
                yield SlotReference(
                    text=token,
                    slot_name=lhs,
                    substitution=Substitutable.parse_substitution(rhs),
                )
            else:
                # Slot without substitutions
                yield SlotReference(text=token, slot_name=token[1:])
        elif ":" in token or "!" in token:
            word = Word(text=token)

            if "!" in token:
                # Word with converter(s)
                # e.g., twenty:20!int
                parts = token.split("!")
                word.text = parts[0]
                word.converters = parts[1:]

            if ":" in word.text:
                # Word with substitution
                # e.g., twenty:20
                lhs, rhs = word.text.split(":", maxsplit=1)
                word.text = lhs
                word.substitution = Substitutable.parse_substitution(rhs)

            yield word
        else:
            # With without substitution
            yield Word(text=token)


def parse_expression(
    root: typing.Optional[Sequence],
    text: str,
    end: typing.List[str] = None,
    is_literal: bool = True,
    metadata: typing.Optional[ParseMetadata] = None,
) -> typing.Optional[int]:
    """Parse a full expression. Return index in text where current expression ends."""
    end = end or []
    found: bool = False
    next_index: int = 0
    literal: str = ""
    last_taggable: typing.Optional[Taggable] = None
    last_group: typing.Optional[Sequence] = root

    # Process text character-by-character
    for current_index, c in enumerate(text):
        if current_index < next_index:
            # Skip ahread
            current_index += 1
            continue

        # Get previous character
        if current_index > 0:
            last_c = text[current_index - 1]
        else:
            last_c = ""

        next_index = current_index + 1

        if c in end:
            # Found end character of expression (e.g., ])
            next_index += 1
            found = True
            break

        if (c in {":", "!"}) and (last_c in {")", "]"}):
            # Handle sequence substitution/conversion
            assert isinstance(last_taggable, Substitutable), parse_error(
                f"Expected Substitutable, got {last_taggable}",
                text,
                current_index,
                metadata=metadata,
            )

            # Check for substitution sequence.
            # e.g., (input words):(output words)
            if text[next_index] == "(":
                # Find end of group
                next_end = [")"] + end
                next_seq_sub = True
            else:
                # Find end of word
                next_end = [" "] + end
                next_seq_sub = False

            next_index = parse_expression(
                None,
                text[current_index + 1 :],
                next_end,
                is_literal=False,
                metadata=metadata,
            )

            if next_index is None:
                # End of text
                next_index = len(text) + 1
            else:
                next_index += current_index - 1

            if next_seq_sub:
                # Consume end paren
                next_index += 1
                assert text[next_index - 1] == ")", parse_error(
                    "Missing end parenthesis", text, current_index, metadata=metadata
                )

            if c == ":":
                # Substitution/conversion
                sub_text = text[current_index + 1 : next_index].strip()

                if "!" in sub_text:
                    # Extract converter(s)
                    sub_text, *converters = sub_text.split("!")
                    last_taggable.converters = converters

                last_taggable.substitution = Substitutable.parse_substitution(sub_text)
            else:
                # Conversion only
                conv_text = text[current_index + 1 : next_index].strip()
                last_taggable.converters = conv_text.split("!")

        elif (c == "(" and last_c != ":") or (c in {"<", "[", "{", "|"}):
            # Begin group/tag/alt/etc.

            # Break literal here
            literal = literal.strip()
            if literal:
                assert last_group is not None, parse_error(
                    "No group preceeding literal",
                    text,
                    current_index,
                    metadata=metadata,
                )
                words = list(split_words(literal))
                last_group.items.extend(words)

                last_word = words[-1]
                assert isinstance(last_word, Taggable), parse_error(
                    f"Expected Taggable, got {last_word}",
                    text,
                    current_index,
                    metadata=metadata,
                )
                last_taggable = last_word
                literal = ""

            if c == "<":
                # Rule reference
                assert last_group is not None, parse_error(
                    "No group preceeding rule reference",
                    text,
                    current_index,
                    metadata=metadata,
                )
                rule = RuleReference()
                end_index = parse_expression(
                    None,
                    text[current_index + 1 :],
                    end=[">"],
                    is_literal=False,
                    metadata=metadata,
                )
                assert end_index, parse_error(
                    f"Failed to find ending '>'", text, current_index, metadata=metadata
                )
                next_index = end_index + current_index

                rule_name = text[current_index + 1 : next_index - 1]
                if "." in rule_name:
                    # Split by last dot
                    last_dot = rule_name.rindex(".")
                    rule.grammar_name = rule_name[:last_dot]
                    rule.rule_name = rule_name[last_dot + 1 :]
                else:
                    # Use entire name
                    rule.rule_name = rule_name

                    if metadata:
                        # Use intent name for grammar name
                        rule.grammar_name = metadata.intent_name

                rule.text = text[current_index:next_index]
                last_group.items.append(rule)
                last_taggable = rule
            elif c == "(":
                # Group (expression)
                if last_group is not None:
                    # Parse group into sequence.
                    # If last_group is None, we're on the right-hand side of a
                    # ":" and the text will be interpreted as a substitution
                    # instead.
                    group = Sequence(type=SequenceType.GROUP)
                    end_index = parse_expression(
                        group, text[current_index + 1 :], end=[")"], metadata=metadata
                    )
                    assert end_index, parse_error(
                        f"Failed to find ending ')'",
                        text,
                        current_index,
                        metadata=metadata,
                    )
                    next_index = end_index + current_index

                    group.text = text[current_index + 1 : next_index - 1]
                    last_group.items.append(group)
                    last_taggable = group
            elif c == "[":
                # Optional
                # Recurse with group sequence to capture multi-word children.
                optional_seq = Sequence(type=SequenceType.GROUP)
                end_index = parse_expression(
                    optional_seq,
                    text[current_index + 1 :],
                    end=["]"],
                    metadata=metadata,
                )
                assert end_index, parse_error(
                    f"Failed to find ending ']'", text, current_index, metadata=metadata
                )
                next_index = end_index + current_index

                optional = Sequence(type=SequenceType.ALTERNATIVE)
                if optional_seq.items:
                    if (
                        (len(optional_seq.items) == 1)
                        and (not optional_seq.tag)
                        and (not optional_seq.substitution)
                    ):
                        # Unpack inner item
                        inner_item = optional_seq.items[0]

                        optional.items.append(inner_item)
                    elif optional_seq.type == SequenceType.ALTERNATIVE:
                        # Unwrap inner alternative
                        optional.items.extend(optional_seq.items)
                    else:
                        # Keep inner group
                        optional_seq.text = text[current_index + 1 : next_index - 1]

                        optional.items.append(optional_seq)

                # Empty alternative
                optional.items.append(Word(text=""))
                optional.text = text[current_index + 1 : next_index - 1]

                assert last_group is not None, parse_error(
                    "Expected group preceeding optional",
                    text,
                    current_index,
                    metadata=metadata,
                )
                last_group.items.append(optional)
                last_taggable = optional
            elif c == "{":
                assert last_taggable is not None, parse_error(
                    "Expected expression preceeding tag",
                    text,
                    current_index,
                    metadata=metadata,
                )
                tag = Tag()

                # Tag
                end_index = parse_expression(
                    None,
                    text[current_index + 1 :],
                    end=["}"],
                    is_literal=False,
                    metadata=metadata,
                )
                assert end_index, parse_error(
                    f"Failed to find ending '}}'",
                    text,
                    current_index,
                    metadata=metadata,
                )
                next_index = end_index + current_index

                # Exclude {}
                tag.tag_text = text[current_index + 1 : next_index - 1]

                # Handle substitution/converter(s)
                if ":" in tag.tag_text or "!" in tag.tag_text:
                    if "!" in tag.tag_text:
                        # Word with converter(s)
                        # e.g., twenty:20!int
                        parts = tag.tag_text.split("!")
                        tag.tag_text = parts[0]
                        tag.converters = parts[1:]

                    if ":" in tag.tag_text:
                        # Word with substitution
                        # e.g., twenty:20
                        lhs, rhs = tag.tag_text.split(":", maxsplit=1)
                        tag.tag_text = lhs
                        tag.substitution = Substitutable.parse_substitution(rhs)

                last_taggable.tag = tag
            elif c == "|":
                assert root is not None, parse_error(
                    "Unexpected '|' outside of group/alternative",
                    text,
                    current_index,
                    metadata=metadata,
                )
                if root.type != SequenceType.ALTERNATIVE:
                    # Create alternative
                    alternative = Sequence(type=SequenceType.ALTERNATIVE)
                    if len(root.items) == 1:
                        # Add directly
                        alternative.items.append(root.items[0])
                    else:
                        # Wrap in group
                        last_group = Sequence(type=SequenceType.GROUP, items=root.items)
                        alternative.items.append(last_group)

                    # Modify original sequence
                    root.items = [alternative]

                    # Overwrite root
                    root = alternative

                assert last_group is not None, parse_error(
                    "Expected group preceeding alternative",
                    text,
                    current_index,
                    metadata=metadata,
                )
                if not last_group.text:
                    # Fix text
                    last_group.text = " ".join(item.text for item in last_group.items)

                # Create new group for any follow-on expressions
                last_group = Sequence(type=SequenceType.GROUP)

                alternative.items.append(last_group)
        else:
            # Accumulate into current literal
            literal += c

    # End of expression
    current_index = len(text)

    # Break literal
    literal = literal.strip()
    if is_literal and literal:
        assert root is not None, parse_error(
            "Literal outside parent expression", text, current_index, metadata=metadata
        )
        words = list(split_words(literal))
        assert last_group is not None, parse_error(
            "Expected group preceeding literal", text, current_index, metadata=metadata
        )
        last_group.items.extend(words)

    if last_group:
        if not last_group.text:
            # Fix text
            last_group.text = " ".join(item.text for item in last_group.items)

        if len(last_group.items) == 1:
            # Simplify final group
            assert root is not None, parse_error(
                "Group outside parent expression",
                text,
                current_index,
                metadata=metadata,
            )
            root.items[-1] = last_group.items[0]

            # Force text to be fixed
            root.text = ""

    if root and (not root.text):
        # Fix text
        if root.type == SequenceType.ALTERNATIVE:
            # Pipe separated
            root.text = " | ".join(item.text for item in root.items)
        else:
            # Space separated
            root.text = " ".join(item.text for item in root.items)

    if end and (not found):
        # Signal end not found
        return None

    return next_index


def parse_error(
    error: str, text: str, column: int, metadata: typing.Optional[ParseMetadata] = None
) -> str:
    """Generate helpful parsing error if metadata is available."""
    if metadata:
        return f"{error} (text='{text}', file={metadata.file_name}, column={column}, line={metadata.line_number})"

    return error



"""Slot load/parsing utility methods."""
import logging
import subprocess
import typing
from dataclasses import dataclass
from pathlib import Path

from .const import IntentsType, ReplacementsType
from .jsgf import Expression, Rule, Sentence, Sequence, SlotReference, walk_expression

_LOGGER = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


@dataclass
class StaticSlotInfo:
    """Name/path to a static slot text file."""

    name: str
    path: Path


@dataclass
class SlotProgramInfo:
    """Name/path/arguments for a slot program."""

    key: str
    name: str
    path: Path
    args: typing.Optional[typing.List[str]] = None


# -----------------------------------------------------------------------------


def get_slot_replacements(
    sentences: IntentsType,
    slots_dirs: typing.Optional[typing.List[Path]] = None,
    slot_programs_dirs: typing.Optional[typing.List[Path]] = None,
    slot_visitor: typing.Optional[
        typing.Callable[[Expression], typing.Union[bool, Expression]]
    ] = None,
) -> ReplacementsType:
    """Create replacement dictionary for referenced slots."""
    replacements: ReplacementsType = {}
    slots_dirs = slots_dirs or []
    slot_programs_dirs = slot_programs_dirs or []

    # Gather used slot names
    slot_names: typing.Set[str] = set()
    for intent_name in sentences:
        for item in sentences[intent_name]:
            for slot_name in get_slot_names(item):
                slot_names.add(slot_name)

    # Load slot values
    for slot_key in slot_names:
        if slot_key in replacements:
            # Skip already loaded slot
            continue

        # Find slot file/program in file system
        slot_info = find_slot(slot_key, slots_dirs, slot_programs_dirs)
        slot_values: typing.List[Expression] = []

        if isinstance(slot_info, StaticSlotInfo):
            # Parse each non-empty line as a JSGF sentence
            _LOGGER.debug("Loading slot %s from %s", slot_key, str(slot_info.path))
            with open(slot_info.path, "r") as slot_file:
                for line in slot_file:
                    line = line.strip()
                    if line:
                        sentence = Sentence.parse(line)
                        if slot_visitor:
                            walk_expression(sentence, slot_visitor)

                        slot_values.append(sentence)
        elif isinstance(slot_info, SlotProgramInfo):
            # Generate values in place
            slot_command = [str(slot_info.path)] + (slot_info.args or [])
            _LOGGER.debug("Running program for slot %s: %s", slot_key, slot_command)

            # Parse each non-empty line as a JSGF sentence
            has_output = False
            for line in subprocess.check_output(
                slot_command, universal_newlines=True
            ).splitlines():
                line = line.strip()
                if line:
                    has_output = True
                    sentence = Sentence.parse(line)
                    if slot_visitor:
                        walk_expression(sentence, slot_visitor)

                    slot_values.append(sentence)

            assert has_output, f"No output from {slot_command}"
        else:
            _LOGGER.warning(
                "Failed to load file/program for slot %s (tried: %s, %s)",
                slot_key,
                slots_dirs,
                slot_programs_dirs,
            )

        # Replace $slot with sentences
        replacements[f"${slot_key}"] = slot_values

    return replacements


# -----------------------------------------------------------------------------


def get_slot_names(item: typing.Union[Expression, Rule]) -> typing.Iterable[str]:
    """Yield referenced slot names from an expression."""
    if isinstance(item, SlotReference):
        yield item.slot_name
    elif isinstance(item, Sequence):
        for sub_item in item.items:
            for slot_name in get_slot_names(sub_item):
                yield slot_name
    elif isinstance(item, Rule):
        for slot_name in get_slot_names(item.rule_body):
            yield slot_name


def split_slot_args(
    slot_name: str,
) -> typing.Tuple[str, typing.Optional[typing.List[str]]]:
    """Split slot name and arguments out (slot,arg1,arg2,...)"""
    # Check for arguments.
    slot_args: typing.Optional[typing.List[str]] = None

    # Slot name retains argument(s).
    if "," in slot_name:
        slot_name, *slot_args = slot_name.split(",")

    return slot_name, slot_args


# -----------------------------------------------------------------------------


def find_slot(
    slot_key: str, slots_dirs: typing.List[Path], slot_programs_dirs: typing.List[Path]
) -> typing.Optional[typing.Union[StaticSlotInfo, SlotProgramInfo]]:
    """Look up a static slot or slot program."""
    # Try static user slots
    for slots_dir in slots_dirs:
        slot_path = slots_dir / slot_key
        if slot_path.is_file():
            return StaticSlotInfo(name=slot_key, path=slot_path)

    # Try user slot programs
    slot_name, slot_args = split_slot_args(slot_key)
    for slot_programs_dir in slot_programs_dirs:
        slot_path = slot_programs_dir / slot_name
        if slot_path.is_file():
            return SlotProgramInfo(
                key=slot_key, name=slot_name, path=slot_path, args=slot_args
            )

    return None



"""Test cases for JSGF graph functions."""
import unittest

from rhasspynlu.ini_jsgf import parse_ini
from rhasspynlu.jsgf_graph import (
    GraphFst,
    GraphFsts,
    graph_to_fst,
    graph_to_fsts,
    intents_to_graph,
)


class FstTestCase(unittest.TestCase):
    """Test cases for OpenFST conversion."""

    def test_single_sentence(self):
        """Test one intent, one sentence."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test
        """
        )

        graph = intents_to_graph(intents)
        fsts = graph_to_fsts(graph)
        self.assertEqual(
            fsts,
            GraphFsts(
                intent_fsts={
                    "TestIntent": "0 1 this this 0\n"
                    "1 2 is is 0\n"
                    "2 3 a a 0\n"
                    "3 4 test test 0\n"
                    "4 5 <eps> <eps> 0\n"
                    "5\n"
                },
                symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
                input_symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
                output_symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
            ),
        )

    def test_substitution(self):
        """Test one intent, one sentence with a substitution."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test:sub
        """
        )

        graph = intents_to_graph(intents)
        fsts = graph_to_fsts(graph)
        self.assertEqual(
            fsts,
            GraphFsts(
                intent_fsts={
                    "TestIntent": "0 1 this this 0\n"
                    "1 2 is is 0\n"
                    "2 3 a a 0\n"
                    "3 4 test <eps> 0\n"
                    "4 5 <eps> sub 0\n"
                    "5 6 <eps> <eps> 0\n"
                    "6\n"
                },
                symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4, "sub": 5},
                input_symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
                output_symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "sub": 5},
            ),
        )

    def test_optional(self):
        """Test one intent, one sentence with an optional word."""
        intents = parse_ini(
            """
        [TestIntent]
        this is [a] test
        """
        )

        graph = intents_to_graph(intents)
        fsts = graph_to_fsts(graph)
        self.assertEqual(
            fsts,
            GraphFsts(
                intent_fsts={
                    "TestIntent": "0 1 this this 0\n"
                    "1 2 is is 0\n"
                    "2 3 a a 0\n"
                    "2 4 <eps> <eps> 0\n"
                    "3 5 <eps> <eps> 0\n"
                    "4 5 <eps> <eps> 0\n"
                    "5 6 test test 0\n"
                    "6 7 <eps> <eps> 0\n"
                    "7\n"
                },
                symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
                input_symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
                output_symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
            ),
        )

    def test_multiple_sentences(self):
        """Test multiple intents."""
        intents = parse_ini(
            """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is another test
        """
        )

        graph = intents_to_graph(intents)
        fsts = graph_to_fsts(graph)
        self.assertEqual(
            fsts,
            GraphFsts(
                intent_fsts={
                    "TestIntent1": "0 1 this this 0\n"
                    "1 2 is is 0\n"
                    "2 3 a a 0\n"
                    "3 4 test test 0\n"
                    "4 5 <eps> <eps> 0\n"
                    "5\n",
                    "TestIntent2": "0 1 this this 0\n"
                    "1 2 is is 0\n"
                    "2 3 another another 0\n"
                    "3 4 test test 0\n"
                    "4 5 <eps> <eps> 0\n"
                    "5\n",
                },
                symbols={
                    "<eps>": 0,
                    "this": 1,
                    "is": 2,
                    "a": 3,
                    "test": 4,
                    "another": 5,
                },
                input_symbols={
                    "<eps>": 0,
                    "this": 1,
                    "is": 2,
                    "a": 3,
                    "test": 4,
                    "another": 5,
                },
                output_symbols={
                    "<eps>": 0,
                    "this": 1,
                    "is": 2,
                    "a": 3,
                    "test": 4,
                    "another": 5,
                },
            ),
        )

    # -------------------------------------------------------------------------

    def test_one_weight(self):
        """Single intent should have an edge weight of 0."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test
        """
        )

        graph = intents_to_graph(intents)
        fst = graph_to_fst(graph)
        self.assertEqual(
            fst,
            GraphFst(
                intent_fst="0 1 <eps> __label__TestIntent 0\n"
                "1 2 this this 0\n"
                "2 3 is is 0\n"
                "3 4 a a 0\n"
                "4 5 test test 0\n"
                "5 6 <eps> <eps> 0\n"
                "6\n",
                symbols={
                    "<eps>": 0,
                    "__label__TestIntent": 1,
                    "this": 2,
                    "is": 3,
                    "a": 4,
                    "test": 5,
                },
                input_symbols={"<eps>": 0, "this": 2, "is": 3, "a": 4, "test": 5},
                output_symbols={
                    "<eps>": 0,
                    "__label__TestIntent": 1,
                    "this": 2,
                    "is": 3,
                    "a": 4,
                    "test": 5,
                },
            ),
        )

    def test_multiple_weights(self):
        """Multiple intents should have balanced weights."""
        intents = parse_ini(
            """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is a test
        """
        )

        graph = intents_to_graph(intents)
        fst = graph_to_fst(graph)
        print(fst)
        self.assertEqual(
            fst,
            GraphFst(
                intent_fst="0 1 <eps> __label__TestIntent1 0.5\n"
                "1 2 this this 0\n"
                "2 3 is is 0\n"
                "3 4 a a 0\n"
                "4 5 test test 0\n"
                "5 6 <eps> <eps> 0\n"
                "0 7 <eps> __label__TestIntent2 0.5\n"
                "7 8 this this 0\n"
                "8 9 is is 0\n"
                "9 10 a a 0\n"
                "10 11 test test 0\n"
                "11 6 <eps> <eps> 0\n"
                "6\n",
                symbols={
                    "<eps>": 0,
                    "__label__TestIntent1": 1,
                    "this": 2,
                    "is": 3,
                    "a": 4,
                    "test": 5,
                    "__label__TestIntent2": 6,
                },
                input_symbols={"<eps>": 0, "this": 2, "is": 3, "a": 4, "test": 5},
                output_symbols={
                    "<eps>": 0,
                    "__label__TestIntent1": 1,
                    "this": 2,
                    "is": 3,
                    "a": 4,
                    "test": 5,
                    "__label__TestIntent2": 6,
                },
            ),
        )

    # -------------------------------------------------------------------------

    def test_intent_filter_single_fst(self):
        """Test multiple intents, single FST with an intent filter."""
        intents = parse_ini(
            """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is another test
        """
        )

        graph = intents_to_graph(intents)
        fst = graph_to_fst(graph, intent_filter=lambda intent: intent == "TestIntent1")
        print(fst)
        self.assertEqual(
            fst,
            GraphFst(
                intent_fst="0 1 <eps> __label__TestIntent1 0.5\n"
                "1 2 this this 0\n"
                "2 3 is is 0\n"
                "3 4 a a 0\n"
                "4 5 test test 0\n"
                "5 6 <eps> <eps> 0\n"
                "6\n",
                symbols={
                    "<eps>": 0,
                    "__label__TestIntent1": 1,
                    "this": 2,
                    "is": 3,
                    "a": 4,
                    "test": 5,
                },
                input_symbols={"<eps>": 0, "this": 2, "is": 3, "a": 4, "test": 5},
                output_symbols={
                    "<eps>": 0,
                    "__label__TestIntent1": 1,
                    "this": 2,
                    "is": 3,
                    "a": 4,
                    "test": 5,
                },
            ),
        )

    def test_intent_filter_multiple_fsts(self):
        """Test multiple intents, multiple FSTs with an intent filter."""
        intents = parse_ini(
            """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is another test
        """
        )

        graph = intents_to_graph(intents)
        fsts = graph_to_fsts(
            graph, intent_filter=lambda intent: intent == "TestIntent1"
        )
        self.assertEqual(
            fsts,
            GraphFsts(
                intent_fsts={
                    "TestIntent1": "0 1 this this 0\n"
                    "1 2 is is 0\n"
                    "2 3 a a 0\n"
                    "3 4 test test 0\n"
                    "4 5 <eps> <eps> 0\n"
                    "5\n"
                },
                symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
                input_symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
                output_symbols={"<eps>": 0, "this": 1, "is": 2, "a": 3, "test": 4},
            ),
        )


# -----------------------------------------------------------------------------


class MiscellaneousTestCase(unittest.TestCase):
    """Miscellaneous JSGF graph test cases"""

    # pylint: disable=R0201
    def test_nested_remote_rule(self):
        """Test a nested rule reference from a separate grammar."""
        intents = parse_ini(
            """
        [TestIntent1]
        test_rule_1 = <test_rule_2>
        test_rule_2 = test
        this is a test

        [TestIntent2]
        this is another <TestIntent1.test_rule_1>
        """
        )

        # Will fail to parse if nested rule references are broken
        intents_to_graph(intents)



"""Test cases for grapheme to phoneme."""
import io
import unittest

from rhasspynlu.g2p import load_sounds_like, load_g2p_corpus, read_pronunciations

_DICTIONARY = """
yawn Y AO N
test T EH S T
say S EY
who HH UW
bee B IY
azure AE ZH ER
read R EH D
read(2) R IY D
"""

_ALIGNMENT = """
a}AE z}ZH u|r}ER e}_
t}T e}EH s}S t}T
"""


class SoundsLikeTests(unittest.TestCase):
    """Test cases for sounds like pronunciations."""

    def setUp(self):
        """Set up tests"""
        with io.StringIO(_DICTIONARY) as dict_file:
            self.pronunciations = read_pronunciations(dict_file)

        with io.StringIO(_ALIGNMENT) as corpus_file:
            self.g2p_alignment = load_g2p_corpus(corpus_file)

    def test_known_words(self):
        """Test pronunciation from known words."""
        sounds_like = """
        beyoncé bee yawn say
        """

        with io.StringIO(sounds_like) as sounds_like_file:
            load_sounds_like(sounds_like_file, self.pronunciations)

        self.assertIn("beyoncé", self.pronunciations)

        # Verify pronunciation is the combination of known word phonemes
        phonemes = self.pronunciations["beyoncé"]
        self.assertEqual(len(phonemes), 1)
        self.assertEqual(phonemes[0], "B IY Y AO N S EY".split())

    def test_homonyms(self):
        """Test word with multiple pronunciations."""
        sounds_like = """
        readbee read bee
        """

        with io.StringIO(sounds_like) as sounds_like_file:
            load_sounds_like(sounds_like_file, self.pronunciations)

        self.assertIn("readbee", self.pronunciations)

        # Verify multiple pronunciations
        phonemes = self.pronunciations["readbee"]
        self.assertEqual(len(phonemes), 2)

        phoneme_strs = [" ".join(p) for p in phonemes]
        self.assertIn("R EH D B IY", phoneme_strs)
        self.assertIn("R IY D B IY", phoneme_strs)

    def test_choose_word(self):
        """Test choosing word from multiple pronunciations."""
        sounds_like = """
        readbee read(1) bee
        """

        with io.StringIO(sounds_like) as sounds_like_file:
            load_sounds_like(sounds_like_file, self.pronunciations)

        self.assertIn("readbee", self.pronunciations)

        # Verify single pronunciation
        phonemes = self.pronunciations["readbee"]
        self.assertEqual(len(phonemes), 1)
        self.assertEqual(phonemes[0], "R EH D B IY".split())

    def test_use_phonemes(self):
        """Test use of literal phonemes."""
        sounds_like = """
        hooiser who /ZH ER/
        """

        with io.StringIO(sounds_like) as sounds_like_file:
            load_sounds_like(sounds_like_file, self.pronunciations)

        self.assertIn("hooiser", self.pronunciations)

        # Verify combination of known word pronunciaton and literal phonemes
        phonemes = self.pronunciations["hooiser"]
        self.assertEqual(len(phonemes), 1)
        self.assertEqual(phonemes[0], "HH UW ZH ER".split())

    def test_word_segment(self):
        """Test use of word segment."""
        sounds_like = """
        hooiser who a>zure<
        """

        with io.StringIO(sounds_like) as sounds_like_file:
            load_sounds_like(
                sounds_like_file, self.pronunciations, g2p_alignment=self.g2p_alignment
            )

        self.assertIn("hooiser", self.pronunciations)

        # Verify combination of known word and word segment
        phonemes = self.pronunciations["hooiser"]
        self.assertEqual(len(phonemes), 1)
        self.assertEqual(phonemes[0], "HH UW ZH ER".split())

    def test_everything(self):
        """Test use of known words, literal phonemes, and word segments."""
        sounds_like = """
        hooiserreadboo who a>zure< read(2) /B UW/
        """

        with io.StringIO(sounds_like) as sounds_like_file:
            load_sounds_like(
                sounds_like_file, self.pronunciations, g2p_alignment=self.g2p_alignment
            )

        self.assertIn("hooiserreadboo", self.pronunciations)

        # Verify combination of known word, literal phonemes, and word segment
        phonemes = self.pronunciations["hooiserreadboo"]
        self.assertEqual(len(phonemes), 1)
        self.assertEqual(phonemes[0], "HH UW ZH ER R IY D B UW".split())


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



"""Methods for computing ngram counts"""
import copy
import itertools
import typing
from collections import Counter, defaultdict, deque

import networkx as nx

from .jsgf_graph import get_start_end_nodes, lcm


def get_intent_ngram_counts(
    graph: nx.DiGraph,
    pad_start="<s>",
    pad_end="</s>",
    order=3,
    balance_counts: bool = True,
) -> typing.Dict[str, Counter]:
    """Gets ngram counts per intent for a JSGF graph"""
    intent_counts: typing.Dict[str, Counter] = defaultdict(Counter)
    start_node, end_node = get_start_end_nodes(graph)
    assert (start_node is not None) and (end_node is not None)

    word_graph = to_word_graph(
        graph, start_node, end_node, pad_start=pad_start, pad_end=pad_end
    )

    sentence_counts: typing.Dict[str, int] = {}

    intent_nodes = set(graph.successors(start_node))
    for intent_node in intent_nodes:
        # __label__INTENT
        edge_data = graph.edges[(start_node, intent_node)]
        olabel = edge_data["olabel"]
        assert olabel.startswith("__label__"), "Not an intent graph"
        intent_name = olabel[9:]
        sentence_counts[intent_name] = edge_data.get("sentence_count", 1)

        # First word(s) of intent
        valid_nodes = set([start_node])
        for maybe_word_node in nx.descendants(graph, intent_node):
            if maybe_word_node in word_graph:
                valid_nodes.add(maybe_word_node)

        # Filter out nodes not part of this intent
        def filter_node(n):
            # pylint: disable=W0640
            return n in valid_nodes

        # Compute ngram counts using a view of the main word graph restricted to
        # nodes from this intent.
        subgraph = nx.subgraph_view(word_graph, filter_node=filter_node)
        intent_counts[intent_name] = count_ngrams(
            subgraph,
            start_node,
            end_node,
            pad_start=pad_start,
            pad_end=pad_end,
            order=order,
        )

    # Balance all n-gram counts by intent
    if balance_counts:
        num_sentences_lcm = lcm(*sentence_counts.values())

        # Multiple all counts by LCM/count for each intent
        for intent_name, sentence_count in sentence_counts.items():
            multiplier = num_sentences_lcm // sentence_count
            ngram_counts = intent_counts[intent_name]
            for ngram in ngram_counts:
                ngram_counts[ngram] *= multiplier

    return intent_counts


# -----------------------------------------------------------------------------


def count_ngrams(
    word_graph: nx.DiGraph,
    start_node: int,
    end_node: int,
    pad_start="<s>",
    pad_end="</s>",
    label="word",
    order=3,
) -> Counter:
    """Compute n-gram counts in a word graph."""
    assert order > 0, "Order must be greater than zero"
    n_data = word_graph.nodes(data=True)

    # Counts from a node to <s>
    up_counts: Counter = Counter()

    # Counts from a node to </s>
    down_counts: Counter = Counter()

    # Top/bottom = 1
    up_counts[start_node] = 1
    down_counts[end_node] = 1

    # Skip start node
    for n in itertools.islice(nx.topological_sort(word_graph), 1, None):
        for n2 in word_graph.predecessors(n):
            up_counts[n] += up_counts[n2]

    # Down
    reverse_graph = nx.reverse_view(word_graph)

    # Skip end node
    for n in itertools.islice(nx.topological_sort(reverse_graph), 1, None):
        for n2 in reverse_graph.predecessors(n):
            down_counts[n] += down_counts[n2]

    # Compute counts
    ngram_counts: Counter = Counter()
    for n in word_graph:
        # Unigram
        word = n_data[n][label]
        ngram = [word]
        ngram_counts[tuple(ngram)] += up_counts[n] * down_counts[n]

        if order == 1:
            continue

        # Higher order
        q = deque([(n, ngram)])
        while q:
            current_node, current_ngram = q.popleft()
            for n2 in word_graph.predecessors(current_node):
                word_n2 = n_data[n2][label]
                ngram_n2 = [word_n2] + current_ngram
                ngram_counts[tuple(ngram_n2)] += up_counts[n2] * down_counts[n]

                if len(ngram_n2) < order:
                    q.append((n2, ngram_n2))

    return ngram_counts


# -----------------------------------------------------------------------------


def to_word_graph(
    graph: nx.DiGraph,
    start_node: int,
    end_node: int,
    pad_start: str = "<s>",
    pad_end: str = "</s>",
    label: str = "word",
) -> nx.DiGraph:
    """Converts a JSGF graph with meta nodes to just words."""

    # Deep copy graph to avoid mutating the original
    graph = copy.deepcopy(graph)
    n_data = graph.nodes(data=True)

    # Add <s> and </s>
    n_data[start_node][label] = pad_start
    n_data[end_node][label] = pad_end

    for node in list(graph):
        word = n_data[node].get(label, "")
        if not word:
            # Clip meta (non-word) node
            for pred_node in graph.predecessors(node):
                for succ_node in graph.successors(node):
                    graph.add_edge(pred_node, succ_node)

            graph.remove_node(node)

    return graph



"""Test cases for ini/JSGF grammar parser."""
import unittest

from rhasspynlu.ini_jsgf import get_intent_counts, parse_ini, split_rules
from rhasspynlu.jsgf import Sentence, Sequence, SequenceType, Tag, Word, walk_expression


class IniJsgfTestCase(unittest.TestCase):
    """Test cases for ini/JSGF grammar parser."""

    def test_parse(self):
        """Test ini/JSGF parser."""
        ini_text = """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is another test
        """

        intents = parse_ini(ini_text)
        self.assertEqual(
            intents,
            {
                "TestIntent1": [
                    Sentence(
                        text="this is a test",
                        items=[Word("this"), Word("is"), Word("a"), Word("test")],
                    )
                ],
                "TestIntent2": [
                    Sentence(
                        text="this is another test",
                        items=[Word("this"), Word("is"), Word("another"), Word("test")],
                    )
                ],
            },
        )

    def test_escape(self):
        """Test escaped optional."""
        ini_text = """
        [TestIntent1]
        \\[this] is a test
        """

        intents = parse_ini(ini_text)
        self.assertEqual(
            intents,
            {
                "TestIntent1": [
                    Sentence(
                        text="[this] is a test",
                        items=[
                            Sequence(
                                text="this",
                                type=SequenceType.ALTERNATIVE,
                                items=[Word("this"), Word("")],
                            ),
                            Word("is"),
                            Word("a"),
                            Word("test"),
                        ],
                    )
                ]
            },
        )

    def test_intent_counts(self):
        """Test sentence counts by intent."""
        ini_text = """
        [TestIntent1]
        this [is] [a] test
        this is [another] test

        [TestIntent2]
        this is (my | your| another) test
        """

        intents = parse_ini(ini_text)
        intent_counts = get_intent_counts(intents)
        self.assertEqual(
            intent_counts,
            {
                "TestIntent1": (1 * 2 * 2 * 1) + (1 * 1 * 2 * 1),
                "TestIntent2": (1 * 1 * 3 * 1),
            },
        )

    def test_transform(self):
        """Test sentence transform."""
        ini_text = """
        [TestIntent1]
        THIS IS A TEST
        """

        intents = parse_ini(ini_text, sentence_transform=str.lower)
        self.assertEqual(
            intents,
            {
                "TestIntent1": [
                    Sentence(
                        text="this is a test",
                        items=[Word("this"), Word("is"), Word("a"), Word("test")],
                    )
                ]
            },
        )

    def test_intent_filter(self):
        """Test filtering intents."""
        ini_text = """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is another test
        """

        intents = parse_ini(ini_text, intent_filter=lambda n: n != "TestIntent2")
        self.assertEqual(
            intents,
            {
                "TestIntent1": [
                    Sentence(
                        text="this is a test",
                        items=[Word("this"), Word("is"), Word("a"), Word("test")],
                    )
                ]
            },
        )

    def test_walk(self):
        """Test Expression.walk with rule and slot reference."""
        ini_text = """
        [SetAlarm]
        minutes = $minute minutes
        set alarm for <minutes>
        """

        intents = parse_ini(ini_text)
        sentences, replacements = split_rules(intents)
        replacements["$minute"] = [Sentence.parse("2 | 3")]

        def num2words(word):
            if not isinstance(word, Word):
                return

            try:
                n = int(word.text)
                if n == 2:
                    word.text = "two"
                    word.substitution = "2"
                elif n == 3:
                    word.text = "three"
                    word.substitution = "3"
            except ValueError:
                pass

        for s in sentences["SetAlarm"]:
            walk_expression(s, num2words, replacements)

        # Verify minute digits were replaced
        minute = replacements["$minute"][0]
        self.assertEqual(
            minute,
            Sentence(
                text="2 | 3",
                type=SequenceType.GROUP,
                items=[
                    Sequence(
                        text="2 | 3",
                        type=SequenceType.ALTERNATIVE,
                        items=[
                            Word("two", substitution="2"),
                            Word("three", substitution="3"),
                        ],
                    )
                ],
            ),
        )

    def test_final_optional_entity(self):
        """Ensure final optional entity has tag."""

        ini_text = """
        [ChangeDisplay]
        display [(page | layer){layout}]
        """

        intents = parse_ini(ini_text)
        self.assertEqual(
            intents,
            {
                "ChangeDisplay": [
                    Sentence(
                        text="display [(page | layer){layout}]",
                        items=[
                            Word("display"),
                            Sequence(
                                text="(page | layer){layout}",
                                type=SequenceType.ALTERNATIVE,
                                items=[
                                    Sequence(
                                        text="page | layer",
                                        type=SequenceType.GROUP,
                                        tag=Tag(tag_text="layout"),
                                        items=[
                                            Sequence(
                                                text="page | layer",
                                                type=SequenceType.ALTERNATIVE,
                                                items=[Word("page"), Word("layer")],
                                            )
                                        ],
                                    ),
                                    Word(""),
                                ],
                            ),
                        ],
                    )
                ]
            },
        )

    def test_rule_with_same_name(self):
        """Ensure referenced rule names don't conflict with local names."""

        ini_text = """
        [Test1]
        rule = is a test
        this <rule>

        [Test2]
        rule = <Test1.rule>
        this <rule>
        """

        _intents = parse_ini(ini_text)
        intents, replacements = split_rules(_intents)

        # Walk all sentences to make sure there's no infinite recursion
        for intent, sentences in intents.items():
            for sentence in sentences:
                walk_expression(sentence, lambda x: x, replacements)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



"""Grapheme to phoneme functions for word pronunciations."""
import itertools
import logging
import re
import shutil
import subprocess
import tempfile
import typing
from collections import defaultdict
from enum import Enum
from pathlib import Path

PronunciationsType = typing.Dict[str, typing.List[typing.List[str]]]

_LOGGER = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class MissingWordPronunciationsException(Exception):
    """Raised when missing word pronunciations and no g2p model."""

    def __init__(self, words: typing.List[str]):
        super().__init__(self)
        self.words = words

    def __str__(self):
        return f"Missing pronunciations for: {self.words}"


# -----------------------------------------------------------------------------


class PronunciationAction(str, Enum):
    """Action taken when multiple pronunciations for the same word are found."""

    APPEND = "append"
    OVERWRITE_ONCE = "overwrite_once"
    OVERWRITE_ALWAYS = "overwrite_always"


def read_pronunciations(
    dict_file: typing.Iterable[str],
    word_dict: typing.Optional[PronunciationsType] = None,
    action: PronunciationAction = PronunciationAction.APPEND,
) -> PronunciationsType:
    """Loads a CMU-like pronunciation dictionary, optionally into an existing dictionary."""
    word_actions: typing.Dict[str, PronunciationAction] = {}

    if word_dict is None:
        word_dict = {}

    for i, line in enumerate(dict_file):
        line = line.strip()
        if not line:
            continue

        try:
            # Use explicit whitespace (avoid 0xA0)
            word, *pronounce = re.split(r"[ \t]+", line)

            word = word.split("(")[0]
            has_word = word in word_dict
            word_action = word_actions.get(word, action)

            if has_word and (word_action == PronunciationAction.APPEND):
                # Append to list of pronunciations
                word_dict[word].append(pronounce)
            elif word_action == PronunciationAction.OVERWRITE_ONCE:
                # Overwrite just once, then append
                word_dict[word] = [pronounce]
                word_actions[word] = PronunciationAction.APPEND
            else:
                # Overwrite
                word_dict[word] = [pronounce]
        except Exception as e:
            _LOGGER.warning("read_pronunciations: %s (line %s)", e, i + 1)

    return word_dict


def write_pronunciations(
    vocabulary: typing.Set[str],
    pronunciations: PronunciationsType,
    dictionary: typing.Union[str, Path],
    g2p_model: typing.Optional[typing.Union[str, Path]] = None,
    g2p_word_transform: typing.Optional[typing.Callable[[str], str]] = None,
    phonetisaurus_apply: typing.Optional[typing.Union[str, Path]] = None,
    missing_words_path: typing.Optional[typing.Union[str, Path]] = None,
    number_repeated_words: bool = True,
):
    """Create pronunciation dictionary. Guess missing words if g2p model is available."""
    # Look up words
    missing_words: typing.Set[str] = set()

    # Look up each word
    with open(dictionary, "w") as dictionary_file:
        for word in vocabulary:
            word_phonemes = pronunciations.get(word)
            if not word_phonemes:
                # Add to missing word list
                _LOGGER.warning("Missing word '%s'", word)
                missing_words.add(word)
                continue

            # Write CMU format
            for i, phonemes in enumerate(word_phonemes):
                phoneme_str = " ".join(phonemes).strip()
                if (not number_repeated_words) or (i == 0):
                    # word
                    print(word, phoneme_str, file=dictionary_file)
                else:
                    # word(n)
                    print(f"{word}({i+1})", phoneme_str, file=dictionary_file)

        # Open missing words file
        missing_file: typing.Optional[typing.TextIO] = None
        if missing_words_path:
            missing_file = open(missing_words_path, "w")

        try:
            if missing_words:
                # Fail if no g2p model is available
                if not g2p_model:
                    raise MissingWordPronunciationsException(list(missing_words))

                if not phonetisaurus_apply:
                    # Find in PATH
                    phonetisaurus_apply = shutil.which("phonetisaurus-apply")
                    assert phonetisaurus_apply, "phonetisaurus-apply not found in PATH"

                # Guess word pronunciations
                _LOGGER.debug("Guessing pronunciations for %s", missing_words)
                guesses = guess_pronunciations(
                    missing_words,
                    g2p_model,
                    phonetisaurus_apply,
                    g2p_word_transform=g2p_word_transform,
                    num_guesses=1,
                )

                # Output is a pronunciation dictionary.
                # Append to existing dictionary file.
                for guess_word, guess_phonemes in guesses:
                    guess_phoneme_str = " ".join(guess_phonemes).strip()
                    print(guess_word, guess_phoneme_str, file=dictionary_file)

                    if missing_file:
                        print(guess_word, guess_phoneme_str, file=missing_file)

        finally:
            if missing_file:
                missing_file.close()
                _LOGGER.debug("Wrote missing words to %s", str(missing_words_path))


def guess_pronunciations(
    words: typing.Iterable[str],
    g2p_model: typing.Union[str, Path],
    phonetisaurus_apply: typing.Optional[typing.Union[str, Path]] = None,
    g2p_word_transform: typing.Optional[typing.Callable[[str], str]] = None,
    num_guesses: int = 1,
) -> typing.Iterable[typing.Tuple[str, typing.List[str]]]:
    """Guess phonetic pronunciations for words. Yields (word, phonemes) pairs."""
    if not phonetisaurus_apply:
        # Find in PATH
        phonetisaurus_apply = shutil.which("phonetisaurus-apply")
        assert phonetisaurus_apply, "phonetisaurus-apply not found in PATH"

    g2p_word_transform = g2p_word_transform or (lambda s: s)

    with tempfile.NamedTemporaryFile(mode="w") as wordlist_file:
        for word in words:
            word = g2p_word_transform(word)
            print(word, file=wordlist_file)

        wordlist_file.seek(0)
        g2p_command = [
            str(phonetisaurus_apply),
            "--model",
            str(g2p_model),
            "--word_list",
            wordlist_file.name,
            "--nbest",
            str(num_guesses),
        ]

        _LOGGER.debug(g2p_command)
        g2p_lines = subprocess.check_output(
            g2p_command, universal_newlines=True
        ).splitlines()

        # Output is a pronunciation dictionary.
        # Append to existing dictionary file.
        for line in g2p_lines:
            line = line.strip()
            if line:
                word, *phonemes = line.split()
                yield (word.strip(), phonemes)


# -----------------------------------------------------------------------------
# Sounds Like Pronunciations
# -----------------------------------------------------------------------------

G2PAlignmentType = typing.Dict[
    str, typing.List[typing.List[typing.Tuple[typing.List[str], typing.List[str]]]]
]
_SOUNDS_LIKE_WORD_N = re.compile(r"^([^(]+)\(([0-9]+)\)$")
_SOUNDS_LIKE_PARTIAL = re.compile(r"^([^>]*)>([^<]+)<.*$")

# -----------------------------------------------------------------------------


def load_sounds_like(
    sounds_like: typing.Union[str, Path, typing.TextIO],
    pronunciations: PronunciationsType,
    action: PronunciationAction = PronunciationAction.APPEND,
    g2p_alignment: typing.Optional[G2PAlignmentType] = None,
    g2p_corpus: typing.Optional[Path] = None,
) -> typing.Optional[G2PAlignmentType]:
    """Loads file with unknown word pronunciations based on known words."""
    original_action = action

    # word -> [[(["graheme", ...], ["phoneme", ...])], ...]
    g2p_alignment = g2p_alignment or {}

    if isinstance(sounds_like, (str, Path)):
        _LOGGER.debug("Loading sounds-like pronunciations from %s", sounds_like)
        sounds_like_file = open(sounds_like, "r")
    else:
        # TextIO
        sounds_like_file = sounds_like

    # File with <unknown_word> <known_word> [<known_word> ...]
    # Pronunciation is derived from phonemes of known words.
    # Phonemes can be included with the syntax /P1 P2/
    with sounds_like_file:
        for i, line in enumerate(sounds_like_file):
            line = line.strip()
            if not line:
                continue

            try:
                # Restore word action
                action = original_action

                # Parse line of <unknown> <known> [<known> ...]
                unknown_word, *known_words = line.split()
                assert known_words, f"No known words for {unknown_word}"

                # Identify literal phonemes
                in_phoneme = False

                # line -> alternatives -> phoneme sequence
                known_phonemes: typing.List[typing.List[typing.List[str]]] = []

                # ongoing phoneme sequence
                current_phonemes: typing.List[str] = []

                # Process space-separated tokens
                for known_word in known_words:
                    if known_word.startswith("/"):
                        # Begin literal phoneme string
                        # /P1 P2 P3/
                        in_phoneme = True
                        known_word = known_word[1:]
                        current_phonemes = []

                    end_slash = known_word.endswith("/")
                    if end_slash:
                        # End literal phoneme string
                        # /P1 P2 P3/
                        known_word = known_word[:-1]

                    if in_phoneme:
                        # Literal phonemes
                        # P_N of /P1 P2 P3/
                        current_phonemes.append(known_word)
                    else:
                        # Check for >part<ial word
                        partial_match = _SOUNDS_LIKE_PARTIAL.match(known_word)
                        if partial_match:
                            partial_prefix, partial_body = (
                                partial_match.group(1),
                                partial_match.group(2),
                            )

                            if not g2p_alignment:
                                # Need to load g2p alignment corpus
                                assert (
                                    g2p_corpus
                                ), f"No G2P corpus given for partial word: {known_word}"
                                assert (
                                    g2p_corpus.is_file()
                                ), f"Missing G2P corpus for {known_word}: {g2p_corpus}"

                                g2p_alignment = load_g2p_corpus(g2p_corpus)

                            # Align graphemes with phonemes
                            word = re.sub(r"[<>]", "", known_word)
                            aligned_phonemes = get_aligned_phonemes(
                                g2p_alignment, word, partial_prefix, partial_body
                            )

                            # Add all possible alignments (phoneme sequences) as alternatives
                            known_phonemes.append(list(aligned_phonemes))
                        else:
                            # Known word with one or more pronunciations
                            known_prons = get_nth_word(pronunciations, known_word)
                            assert known_prons, f"No pronunciations for {known_word}"

                            # Add all pronunciations as alternatives
                            known_phonemes.append(known_prons)

                    if end_slash:
                        in_phoneme = False
                        if current_phonemes:
                            known_phonemes.append([current_phonemes])

                # Collect pronunciations from known words
                # word_prons: typing.List[typing.List[typing.List[str]]] = []
                for word_phonemes in itertools.product(*known_phonemes):
                    # Generate all possible pronunciations.
                    word_pron = list(itertools.chain(*word_phonemes))
                    has_word = unknown_word in pronunciations

                    # Handle according to custom words action
                    if has_word and (action == PronunciationAction.APPEND):
                        # Append to list of pronunciations
                        pronunciations[unknown_word].append(word_pron)
                    elif action == PronunciationAction.OVERWRITE_ONCE:
                        # Overwrite just once, then append
                        pronunciations[unknown_word] = [word_pron]
                        action = PronunciationAction.APPEND
                    else:
                        # Overwrite
                        pronunciations[unknown_word] = [word_pron]
            except Exception as e:
                _LOGGER.warning("load_sounds_like: %s (line %s)", e, i + 1)
                raise e

    return g2p_alignment


def load_g2p_corpus(
    g2p_corpus: typing.Union[str, typing.TextIO, Path]
) -> G2PAlignmentType:
    """Loads a grapheme to phoneme alignment corpus generated by Phonetisaurus."""
    g2p_alignment: G2PAlignmentType = defaultdict(list)

    if isinstance(g2p_corpus, (str, Path)):
        _LOGGER.debug("Loading g2p corpus from %s", g2p_corpus)
        corpus_file = open(g2p_corpus, "r")
    else:
        # TextIO
        corpus_file = g2p_corpus

    with corpus_file:
        for line in corpus_file:
            line = line.strip()
            if not line:
                continue

            word = ""
            inputs_outputs = []

            # Parse line
            parts = line.split()
            for part in parts:
                # Assume default delimiters:
                # } separates input/output
                # | separates input/output tokens
                # _ indicates empty output
                part_in, part_out = part.split("}")
                part_ins = part_in.split("|")
                if part_out == "_":
                    # Empty output
                    part_outs = []
                else:
                    part_outs = part_out.split("|")

                inputs_outputs.append((part_ins, part_outs))
                word += "".join(part_ins)

            # Add to pronunciations for word
            g2p_alignment[word].append(inputs_outputs)

    return g2p_alignment


def get_aligned_phonemes(
    g2p_alignment: G2PAlignmentType, word: str, prefix: str, body: str
) -> typing.Iterable[typing.List[str]]:
    """Yields lists of phonemes that comprise the body of the word. Prefix graphemes are skipped."""
    word_index: typing.Optional[int] = None
    match = _SOUNDS_LIKE_WORD_N.match(word)
    if match:
        # word(N)
        word, word_index = (match.group(1), int(match.group(2)))

    # Loop through possible alignments for this word
    for io_index, inputs_outputs in enumerate(g2p_alignment.get(word, [])):
        if (word_index is not None) and (word_index != (io_index + 1)):
            continue

        can_match = True
        prefix_chars = list(prefix)
        body_chars = list(body)

        phonemes = []
        for word_input, word_output in inputs_outputs:
            word_input = list(word_input)
            word_output = list(word_output)

            while prefix_chars and word_input:
                # Exhaust characters before desired word segment first
                if word_input[0] != prefix_chars[0]:
                    can_match = False
                    break

                prefix_chars = prefix_chars[1:]
                word_input = word_input[1:]

            while body_chars and word_input:
                # Match desired word segment
                if word_input[0] != body_chars[0]:
                    can_match = False
                    break

                body_chars = body_chars[1:]
                word_input = word_input[1:]

                if word_output:
                    phonemes.append(word_output[0])
                    word_output = word_output[1:]

            if not can_match or not body_chars:
                # Mismatch or done with word segment
                break

        if can_match and phonemes:
            yield phonemes


def get_nth_word(
    pronunciations: PronunciationsType, word: str
) -> typing.List[typing.List[str]]:
    """Get all pronunciations for a word or a single(n) pronunciation."""
    # Check for explicit word index (1-based)
    word_index: typing.Optional[int] = None
    match = _SOUNDS_LIKE_WORD_N.match(word)
    if match:
        # word(N)
        word, word_index = (match.group(1), int(match.group(2)))

    known_prons = pronunciations.get(word, [])
    if (not known_prons) or (word_index is None):
        # Add all known pronunciations
        return known_prons

    # Add indexed word only.
    # Clip to within bounds of list.
    i = min(max(1, word_index), len(known_prons)) - 1
    return [known_prons[i]]
